/**
 * Notera:
 *
 * Om du vill kunna kompilera koden måste du kopiera mappen "wrap"
 * samt "Makefile" från de givna tentafilerna till samma mapp som du
 * lägger denna filen i.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include "wrap/thread.h"
#include "wrap/synch.h"
#include "wrap/atomics.h"
#include "picture.h"

/**
 * Följande datastrukturer finns i picture.h:
 * - struct line - ett linjesegment
 * - struct picture - en bild, består av ett antal linjesegment
 *
 * Följande funktioner finns i picture.h:
 *
 * - picture_load: ladda en bild från en fil. Okej att anropa från
 *   flera trådar samtidigt.
 * - picture_free: förstör en bild.
 * - picture_compute_row: beräkna en rad (en viss y-koordinat)
 *   i bilden. Läser endast från picture-objektet.
 */


// Data som skickas till "picture_worker" och delas mellan trådarna
// som ritar upp bilden.
struct draw_data {
    // Bilden som ska ritas upp.
    struct picture *picture;

    // Nästa rad av pixlar som ska genereras.
    int line_to_create;

    // Nästa rad av pixlar som ska ritas ut.
    int line_to_draw;
};

// Funktion som körs av de trådar som skapas i "draw_picture". Du kan
// anta att denna funktion bara anropas från "draw_picture".
static void draw_worker(void *arg) {
    struct draw_data *data = arg;

    while (true) {
        // Lista ut vilken rad vi ska generera.
        int my_line = atomic_add(&data->line_to_create, 1);
        if (my_line >= data->picture->height)
            break;

        // Skapa raden.
        char *line = picture_compute_row(data->picture, my_line);

        // Vänta tills det är vår tur att rita upp den.
        while (atomic_read(&data->line_to_draw) != my_line)
            ;

        // Skriv ut den till standard output.
        printf("%s\n", line);

        atomic_add(&data->line_to_draw, 1);

        free(line);
    }
}

// Rita upp en bild till standard output.
// Notera: du har planer på att utöka denna funktion så att den i
// framtiden också kan spara bilden till en fil. Du vill därmed
// redan nu synkronisera funktionen så att det går att köra den
// parallellt från flera trådar samtidigt.
void draw_picture(struct picture *p, int num_threads) {
    struct draw_data data;
    data.picture = p;
    data.line_to_create = 0;
    data.line_to_draw = 0;

    for (int i = 0; i < num_threads; i++)
        thread_new(&draw_worker, &data);

    while (atomic_read(&data.line_to_draw) < p->height)
        ;
}


/**
 * Huvudprogram.
 *
 * Eventuella ändringar här nedanför kommer att ignoreras vid rättning
 * av uppgiften. Du får däremot ändra nedanför ifall du vill testa din
 * lösning.
 */

int main(int argc, const char **argv) {
    const char *picture = "pintos.txt";
    if (argc > 1)
        picture = argv[1];

    struct picture *p = picture_load(picture);
    if (p == NULL) {
        printf("No picture named %s!\n", picture);
        return 1;
    }

    draw_picture(p, 4);
    picture_free(p);

    return 0;
}
