/**
 * Notera:
 *
 * Om du vill kunna kompilera koden måste du kopiera mappen "wrap"
 * samt "Makefile" från de givna tentafilerna till samma mapp som du
 * lägger denna filen i.
 *
 * a)
 *
 * Exempelvis: två trådar kör "track_enter", den ena sätter in tåg
 * nummer 1, den andra tåg nummer 2. Båda trådarna kommer till
 * while-loopen och ser att första segmentet är ledigt. I och med att
 * ingen av trådarna har hunnit skriva in sitt tågnummer i
 * "occupied_by" går båda trådarna förbi while-loopen och skriver sitt
 * tågnummer i "occupied_by" i första segmentet. Detta gör att endast
 * det ena tågnumret finns kvar i slutändan.
 *
 * b)
 *
 * Exempelvis: "track_draw" körs av en tråd. Funktionen ritar ut
 * segment 1 till och med 3. I segment 3 finns tåg nummer 8. En annan
 * tråd flyttar tåget från segment 3 till segment 4. Utritningen
 * fortsätter sedan att rita ut segment 4 till 12. I det här fallet
 * kommer tåg 8 att finnas i både segment 3 och segment 4.
 *
 * c)
 *
 * Ja, markerade nedan i "track_enter" och "track_move".
 *
 * d)
 *
 * Se nedan.
 *
 * e)
 *
 * Se modifikationer i koden. Markerade med // (e)
 *
 * f)
 *
 * De fyra villkoren för deadlock är:
 * - Mutual exclusion: det finns resurser i systemet som vi måste
 *   se till att endast en tråd använder åt gången (notera: ibland
 *   kan det finnas flera likvärdiga instanser av dessa resurser).
 * - No preemption: en tråd som använder en resurs måste lämna
 *   tillbaka den frivilligt. Andra trådar kan inte tvinga en tråd
 *   att lämna tillbaka resurser.
 * - Hold and wait: det är möjligt att trådar håller en resurs medan
 *   den väntar på att få tillgång till en annan resurs också.
 *
 * Om villkoren ovan är uppfyllda så kan följande villkor också
 * uppfyllas:
 * - Circular wait: trådar väntar på varandra i en cykel. En tråd
 *   håller i en resurs och väntar på en annan resurs, medan en annan
 *   tråd håller i den andra resursen och väntar på den första.
 *
 * g)
 *
 * De tre första villkoren är uppfyllda:
 * - Mutual exclusion: vi använder lås för att skydda "occupied_by".
 * - No preemption: detta är hur våra lås fungerar, vi har ingen
 *   operation "lock_steal" eller liknande.
 * - Hold and wait: det finns situationer där vi försöker ta två
 *   lås samtidigt.
 *
 * Circular wait uppfylls dock inte:
 * - I "track_move" ser vi till att alltid ta det lägst numrerade
 *   segmentet först för att se till att vi tar låsen i samma ordning
 *   som i "track_draw".
 * - Spåren är enkelriktade, vilket gör att vi inte kan få en cykel
 *   av tåg på ett spår som väntar på varandra med semaforen "available".
 *   Tåget längst fram på spåret kommer alltid att kunna åka vidare.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include "wrap/thread.h"
#include "wrap/synch.h"
#include "wrap/atomics.h"

/**
 * Ett segment av ett spår. För att undvika kollisioner vill vi
 * garantera att det bara finns ett tåg i varje segment.
 */
struct segment {
    // Vilket tåg som finns i segmentet just nu. Noll betyder att
    // segmentet är tomt.
    int occupied_by;

    // Lås för "occupied_by"
    struct lock occupied_lock; // (e)

    // Semafor för att vänta på att segmentet blir ledigt. Räknar lediga platser.
    struct semaphore available; // (e)
};


/**
 * Ett spår. Varje spår består av ett eller flera segment. I varje
 * segment kan maximalt ett tåg befinna sig.
 */
struct track {
    // Array av segment.
    struct segment *segments;

    // Antal segment i "segments".
    int segment_count;

    // Vilken riktning tågen får röra sig. Antingen 1 eller -1.
    int direction;
};


// Skapa ett spår bestående av ett visst antal segment. "direction" är
// antingen 1 eller -1 och indikerar i vilken riktning tågen får åka.
struct track *track_create(int segment_count, int direction) {
    struct track *track = malloc(sizeof(struct track));
    track->segments = malloc(sizeof(struct segment)*segment_count);
    track->segment_count = segment_count;
    track->direction = direction;

    for (int i = 0; i < segment_count; i++) {
        struct segment *segment = &track->segments[i];
        segment->occupied_by = 0;
        lock_init(&segment->occupied_lock); // (e)
        sema_init(&segment->available, 1); // (e)
    }

    return track;
}

// Frigör minnet som används av ett spår.
void track_free(struct track *track) {
    free(track->segments);
    free(track);
}

// Anropas när ett tåg ska börja åka på spåret. Tågen börjar alltid
// åka i början av spåret (vad som är början beror på "direction").
// Returnerar segmentet som tåget börjar i.
int track_enter(struct track *track, int train_id) {
    int segment_id = 0;
    if (track->direction < 0)
        segment_id = track->segment_count - 1;

    struct segment *segment = &track->segments[segment_id];

    sema_down(&segment->available); // (e)

    // Kritisk sektion 1 börjar. Resurs: segment->occupied_by
    lock_acquire(&segment->occupied_lock);

    // Vänta på att tåget får plats.
    // while (segment->occupied_by != 0)    // <--  busy wait
    //     ;
    segment->occupied_by = train_id;

    // Kritisk sektion 1 slutar.
    lock_release(&segment->occupied_lock); // (e)

    return segment_id;
}

// Anropas när ett tåg ska flytta sig från ett segment till ett
// annat. Tåget flyttar sig från segmentet "from_segment" och ett steg
// i riktningen som anges av spårets riktning.
void track_move(struct track *track, int from_segment) {
    struct segment *from = &track->segments[from_segment];
    struct segment *to = &track->segments[from_segment + track->direction];

    sema_down(&to->available); // (e)

    // Kritisk sektion 2 börjar. Resurs: from->occupied_by och to->occupied_by
    if (track->direction > 0) {             // (e)
        lock_acquire(&from->occupied_lock); // (e)
        lock_acquire(&to->occupied_lock);   // (e)
    } else {                                // (e)
        lock_acquire(&to->occupied_lock);   // (e)
        lock_acquire(&from->occupied_lock); // (e)
    }                                       // (e)

    // while (to->occupied_by != 0)      // <-- busy wait
    //     ;
    to->occupied_by = from->occupied_by;
    from->occupied_by = 0;

    // Kritisk sektion 2 slutar.
    lock_release(&from->occupied_lock);  // (e)
    lock_release(&to->occupied_lock);    // (e)

    sema_up(&from->available); // (e)
}

// Anropas när tåget åker av spåret (dvs. när tåget har kommit fram
// till sin destination).
void track_leave(struct track *track, int segment_id) {
    struct segment *segment = &track->segments[segment_id];

    // Kritisk sektion 3 börjar. Resurs: segment->occupied_by
    lock_acquire(&segment->occupied_lock); // (e)

    segment->occupied_by = 0;

    // Kritisk sektion 3 slutar.
    lock_release(&segment->occupied_lock); // (e)

    sema_up(&segment->available); // (e)
}

// Rita ut spåret till standard output. Vi antar att ingen annan tråd
// skriver saker till standard output när denna funktion anropas.
void track_draw(struct track *track) {
    // Ovansidan:
    printf("|");
    for (int i = 0; i < track->segment_count; i++) {
        printf("----|");
    }
    printf("\n");

    // Inuti:
    printf("|");

    // Kritisk sektion 4 börjar. Resurs: occupied_by i *alla* segment.
    for (int i = 0; i < track->segment_count; i++)       // (e)
        lock_acquire(&track->segments[i].occupied_lock); // (e)

    for (int i = 0; i < track->segment_count; i++) {
        int occupant = track->segments[i].occupied_by;
        if (occupant != 0)
            printf("%4d|", occupant);
        else
            printf("   -|");
    }

    // Kritisk sektion 4 slutar.
    for (int i = 0; i < track->segment_count; i++)       // (e)
        lock_release(&track->segments[i].occupied_lock); // (e)

    printf("\n");

    printf("|");
    for (int i = 0; i < track->segment_count; i++) {
        printf("----|");
    }
    printf("\n");
}

/**
 * Huvudprogram.
 *
 * Eventuella ändringar här nedanför kommer att ignoreras vid rättning
 * av uppgiften. Du får däremot ändra nedanför ifall du vill testa din
 * lösning.
 */

// Håll koll på vilka nummer som tågen har fått.
int train_id = 1;
struct lock train_id_lock;

// Håll koll på vilka trådar som är klara.
int trains_done = 0;
struct lock done_lock;

// Hjälpfunktion för att sova en slumpmässig tid.
void random_sleep(void) {
    // Notera: "rand" är säker att anropa från flera trådar samtidigt.
    int time = (rand() % 300) + 500;
    timer_msleep(time);
}

// Kör ett tåg längs ett spår. Anropas från main nedan.
void run_train(void *data) {
    struct track *track = data;

    // Hämta ett unikt ID för tåget.
    lock_acquire(&train_id_lock);
    int id = train_id++;
    lock_release(&train_id_lock);

    // Börja från början.
    int position = track_enter(track, id);
    random_sleep();

    // Stega igenom alla segment.
    for (int i = 1; i < track->segment_count; i++) {
        track_move(track, position);
        position += track->direction;

        random_sleep();
    }

    // Sedan är vi klara!
    track_leave(track, position);

    lock_acquire(&done_lock);
    trains_done++;
    lock_release(&done_lock);
}

int main(void) {
    lock_init(&train_id_lock);
    lock_init(&done_lock);

    struct track *outbound = track_create(10, 1);
    struct track *inbound = track_create(10, -1);

    int threads_to_spawn = 6;

    // Starta trådar som representerar tåg i olika riktningar.
    for (int i = 0; i < threads_to_spawn; i++) {
        thread_new(&run_train, outbound);
        thread_new(&run_train, inbound);
    }

    // Uppdatera skärmen tills alla tåg har åkt till sin destination.
    bool done = false;
    while (!done) {
        lock_acquire(&done_lock);
        if (trains_done == 2*threads_to_spawn)
            done = true;
        lock_release(&done_lock);

        // Rensa terminalen.
        printf("\e[1;1H\e[2J");

        // Skriv ut ny data.
        printf("Outbound:\n");
        track_draw(outbound);
        printf("Inbound:\n");
        track_draw(inbound);

        timer_msleep(100);
    }

    track_free(outbound);
    track_free(inbound);

    return 0;
}
